import logging
from pathlib import Path
from typing import Dict, Iterator, Tuple

import pytest
import requests
from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_fixed

from pglift import exceptions
from pglift import instance as instance_mod
from pglift import prometheus, systemd
from pglift.ctx import Context
from pglift.models import interface, system

from . import reconfigure_instance


def config_dict(configpath: Path) -> Dict[str, str]:
    config = {}
    for line in configpath.read_text().splitlines():
        key, value = line.split("=", 1)
        config[key] = value.strip()
    return config


def test_configure(
    ctx: Context,
    installed: None,
    instance: system.Instance,
    instance_manifest: interface.Instance,
    tmp_port_factory: Iterator[int],
) -> None:
    prometheus_settings = ctx.settings.prometheus
    name = instance.qualname
    configpath = Path(str(prometheus_settings.configpath).format(name=name))
    assert configpath.exists()

    prometheus_config = config_dict(configpath)
    dsn = prometheus_config["DATA_SOURCE_NAME"]
    assert "user=postgres" in dsn
    assert f"port={instance.port}" in dsn
    port = instance.prometheus.port
    assert prometheus_config["PG_EXPORTER_WEB_LISTEN_ADDRESS"] == f":{port}"

    queriespath = Path(str(prometheus_settings.queriespath).format(name=name))
    assert queriespath.exists()

    new_port = next(tmp_port_factory)
    with reconfigure_instance(ctx, instance, instance_manifest, port=new_port):
        new_prometheus_config = config_dict(configpath)
        dsn = new_prometheus_config["DATA_SOURCE_NAME"]
        assert f"port={new_port}" in dsn


@pytest.fixture
def postgres_exporter(
    ctx: Context,
    instance_manifest: interface.Instance,
    instance: system.Instance,
    installed: None,
    tmp_port_factory: Iterator[int],
) -> Iterator[Tuple[str, str, int]]:
    """Setup a postgres_exporter service for 'instance' using another port."""
    port = next(tmp_port_factory)
    name = "123-fo-o"
    role = interface.instance_surole(ctx.settings, instance_manifest)
    dsn = f"dbname=postgres port={instance.port} user={role.name} sslmode=disable"
    host = instance.config().get("unix_socket_directories")
    if host:
        dsn += f" host={host}"
    password = None
    if role.password:
        password = role.password.get_secret_value()
    prometheus.setup(ctx, name, dsn=dsn, password=password, port=port)
    prometheus_settings = ctx.settings.prometheus
    configpath = Path(str(prometheus_settings.configpath).format(name=name))
    assert configpath.exists()
    queriespath = Path(str(prometheus_settings.queriespath).format(name=name))
    assert queriespath.exists()

    yield name, dsn, port

    prometheus.revert_setup(ctx, name)
    assert not configpath.exists()
    assert not queriespath.exists()


def test_setup(
    ctx: Context, instance: system.Instance, postgres_exporter: Tuple[str, str, int]
) -> None:
    name, dsn, port = postgres_exporter
    configpath = Path(str(ctx.settings.prometheus.configpath).format(name=name))

    prometheus_config = config_dict(configpath)
    assert f"port={instance.port}" in prometheus_config["DATA_SOURCE_NAME"]
    assert prometheus_config["PG_EXPORTER_WEB_LISTEN_ADDRESS"] == f":{port}"


@retry(reraise=True, wait=wait_fixed(1), stop=stop_after_attempt(3))
def request_metrics(port: int) -> requests.Response:
    return requests.get(f"http://0.0.0.0:{port}/metrics")


def test_start_stop(ctx: Context, installed: None, instance: system.Instance) -> None:
    port = instance.prometheus.port

    if ctx.settings.service_manager == "systemd":
        assert systemd.is_enabled(ctx, prometheus.systemd_unit(instance.qualname))

    with instance_mod.running(ctx, instance, run_hooks=True):
        if ctx.settings.service_manager == "systemd":
            assert systemd.is_active(ctx, prometheus.systemd_unit(instance.qualname))
        try:
            r = request_metrics(port)
        except requests.ConnectionError as e:
            raise AssertionError(f"HTTP connection failed: {e}") from None
        r.raise_for_status()
        assert r.ok
        output = r.text
        assert "pg_up 1" in output.splitlines()

    with instance_mod.stopped(ctx, instance, run_hooks=True):
        if ctx.settings.service_manager == "systemd":
            assert not systemd.is_active(
                ctx, prometheus.systemd_unit(instance.qualname)
            )
        with pytest.raises(requests.ConnectionError):
            request_metrics(port)


def test_start_stop_nonlocal(
    ctx: Context, instance: system.Instance, postgres_exporter: Tuple[str, str, int]
) -> None:
    name, dsn, port = postgres_exporter

    if ctx.settings.service_manager == "systemd":
        assert systemd.is_enabled(ctx, prometheus.systemd_unit(name))

    with instance_mod.running(ctx, instance, run_hooks=False):
        prometheus.start(ctx, name)
        try:
            if ctx.settings.service_manager == "systemd":
                assert systemd.is_active(ctx, prometheus.systemd_unit(name))
            try:
                r = request_metrics(port)
            except requests.ConnectionError as e:
                raise AssertionError(f"HTTP connection failed: {e}") from None
            r.raise_for_status()
            assert r.ok
            output = r.text
            assert "pg_up 1" in output.splitlines()
        finally:
            prometheus.stop(ctx, name)

        if ctx.settings.service_manager == "systemd":
            assert not systemd.is_active(ctx, prometheus.systemd_unit(name))
        with pytest.raises(requests.ConnectionError):
            request_metrics(port)


def test_apply(ctx: Context, tmp_port_factory: Iterator[int]) -> None:
    port = next(tmp_port_factory)
    m = interface.PostgresExporter(name="test", dsn="dbname=test", port=port)
    prometheus.apply(ctx, m)

    prometheus_settings = ctx.settings.prometheus
    configpath = Path(str(prometheus_settings.configpath).format(name="test"))
    assert configpath.exists()
    queriespath = Path(str(prometheus_settings.queriespath).format(name="test"))
    assert queriespath.exists()

    prometheus_config = config_dict(configpath)
    assert prometheus_config["PG_EXPORTER_WEB_LISTEN_ADDRESS"] == f":{port}"

    port1 = next(tmp_port_factory)
    prometheus.apply(ctx, m.copy(update={"port": port1}))
    prometheus_config = config_dict(configpath)
    assert prometheus_config["PG_EXPORTER_WEB_LISTEN_ADDRESS"] == f":{port1}"

    prometheus.apply(
        ctx, interface.PostgresExporter(name="test", dsn="", port=port, state="absent")
    )
    assert not configpath.exists()
    assert not queriespath.exists()


def test_drop_exists(
    ctx: Context, tmp_port_factory: Iterator[int], caplog: pytest.LogCaptureFixture
) -> None:
    port = next(tmp_port_factory)
    prometheus.setup(ctx, "dropme", port=port)
    assert prometheus.port(ctx, "dropme") == port
    assert prometheus.exists(ctx, "dropme")
    prometheus.drop(ctx, "dropme")
    assert not prometheus.exists(ctx, "dropme")
    with pytest.raises(exceptions.FileNotFoundError, match="postgres_exporter config"):
        prometheus.port(ctx, "dropme")
    caplog.clear()
    with caplog.at_level(logging.WARNING, logger="pglift"):
        prometheus.drop(ctx, "dropme")
    assert caplog.records[0].message == "no postgres_exporter service 'dropme' found"



import argparse
from typing import Optional, Sequence

from . import exceptions
from .cmd import start_program
from .ctx import Context
from .exceptions import InstanceNotFound
from .models.system import PostgreSQLInstance
from .pm import PluginManager

parser = argparse.ArgumentParser(description="Start postgres for specified instance")
parser.add_argument(
    "instance",
    help="instance identifier as <version>-<name>",
)


def main(
    argv: Optional[Sequence[str]] = None,
    *,
    ctx: Optional[Context] = None,
) -> int:
    args = parser.parse_args(argv)
    if ctx is None:
        ctx = Context(plugin_manager=PluginManager.get())

    try:
        instance = PostgreSQLInstance.from_stanza(ctx, args.instance)
    except ValueError as e:
        parser.error(str(e))
    except InstanceNotFound as e:
        parser.exit(2, str(e))

    bindir = ctx.settings.postgresql.versions[instance.version].bindir
    cmd = [str(bindir / "postgres"), "-D", str(instance.datadir)]
    pidfile = (
        ctx.settings.postgresql.pid_directory
        / f"postgresql-{instance.version}-{instance.name}.pid"
    )
    try:
        start_program(cmd, pidfile, capture_output=False)
    except exceptions.CommandError:
        return 1
    else:
        return 0


if __name__ == "__main__":  # pragma: nocover
    import sys

    sys.exit(main())



from datetime import datetime, timezone
from unittest.mock import patch

from pglift import pgbackrest
from pglift.ctx import Context
from pglift.models.system import Instance
from pglift.settings import Settings


def test_make_cmd(pg_version: str, settings: Settings, instance: Instance) -> None:
    assert pgbackrest.make_cmd(instance, settings.pgbackrest, "stanza-upgrade") == [
        "/usr/bin/pgbackrest",
        f"--config={settings.prefix}/etc/pgbackrest/pgbackrest-{pg_version}-test.conf",
        f"--stanza={pg_version}-test",
        "stanza-upgrade",
    ]


def test_backup_info(
    ctx: Context, settings: Settings, pg_version: str, instance: Instance
) -> None:
    with patch.object(ctx, "run") as run:
        run.return_value.stdout = "[]"
        assert pgbackrest.backup_info(ctx, instance, backup_set="foo") == []
    run.assert_called_once_with(
        [
            "/usr/bin/pgbackrest",
            f"--config={settings.prefix}/etc/pgbackrest/pgbackrest-{pg_version}-test.conf",
            f"--stanza={pg_version}-test",
            "--set=foo",
            "--output=json",
            "info",
        ],
        check=True,
    )


def test_backup_command(
    pg_version: str, settings: Settings, instance: Instance
) -> None:
    assert pgbackrest.backup_command(instance, type=pgbackrest.BackupType.full) == [
        "/usr/bin/pgbackrest",
        f"--config={settings.prefix}/etc/pgbackrest/pgbackrest-{pg_version}-test.conf",
        f"--stanza={pg_version}-test",
        "--type=full",
        "--log-level-console=info",
        "--start-fast",
        "backup",
    ]


def test_expire_command(
    pg_version: str, settings: Settings, instance: Instance
) -> None:
    assert pgbackrest.expire_command(instance) == [
        "/usr/bin/pgbackrest",
        f"--config={settings.prefix}/etc/pgbackrest/pgbackrest-{pg_version}-test.conf",
        f"--stanza={pg_version}-test",
        "--log-level-console=info",
        "expire",
    ]


def test_restore_command(
    pg_version: str, settings: Settings, instance: Instance
) -> None:
    assert pgbackrest.restore_command(
        instance, date=datetime(2003, 1, 1).replace(tzinfo=timezone.utc), backup_set="x"
    ) == [
        "/usr/bin/pgbackrest",
        f"--config={settings.prefix}/etc/pgbackrest/pgbackrest-{pg_version}-test.conf",
        f"--stanza={pg_version}-test",
        "--log-level-console=info",
        "--link-all",
        "--target-action=promote",
        "--type=time",
        "--target=2003-01-01 00:00:00.000000+0000",
        "--set=x",
        "restore",
    ]



import builtins
import contextlib
import shutil
import tempfile
import time
from pathlib import Path
from typing import Any, Dict, Iterator, Optional, Tuple, Union

from pgtoolkit import conf as pgconf
from pgtoolkit import pgpass
from pgtoolkit.ctl import Status as Status
from pydantic import SecretStr
from typing_extensions import Literal

from . import (
    cmd,
    conf,
    datapath,
    db,
    exceptions,
    hookimpl,
    roles,
    systemd,
    template,
    util,
)
from .ctx import BaseContext
from .models import interface
from .models.system import (
    BaseInstance,
    Instance,
    PostgreSQLInstance,
    default_postgresql_version,
)
from .task import task
from .types import ConfigChanges


def systemd_unit(instance: BaseInstance) -> str:
    return f"pglift-postgresql@{instance.version}-{instance.name}.service"


def init_replication(
    ctx: BaseContext, instance: BaseInstance, standby_for: str, slot: Optional[str]
) -> None:
    with tempfile.TemporaryDirectory() as _tmpdir:
        tmpdir = Path(_tmpdir)
        # pg_basebackup will also copy config files from primary datadir.
        # So to have expected configuration at this stage we have to backup
        # postgresql.conf & pg_hba.conf (created by prior pg_ctl init) and
        # restore after pg_basebackup finishes.
        keep = {"postgresql.conf", "pg_hba.conf"}
        for name in keep:
            shutil.copyfile(instance.datadir / name, tmpdir / name)
        shutil.rmtree(instance.datadir)
        shutil.rmtree(instance.waldir)
        bindir = ctx.pg_ctl(instance.version).bindir
        cmd = [
            str(bindir / "pg_basebackup"),
            "--pgdata",
            str(instance.datadir),
            "--write-recovery-conf",
            "--checkpoint=fast",
            "--no-password",
            "--progress",
            "--verbose",
            "--dbname",
            standby_for,
            "--waldir",
            str(instance.waldir),
        ]

        env = ctx.libpq_environ()
        if slot:
            cmd += ["--slot", slot]
            with db.connect_dsn(
                standby_for,
                passfile=env.get("PGPASSFILE"),
                password=env.get("PGPASSWORD"),
            ) as cnx:
                # ensure the replication slot does not exists
                # otherwise --create-slot will raise an error
                with cnx.cursor() as cur:
                    cur.execute(db.query("drop_replication_slot"), {"slot": slot})
                    if int(instance.version) <= 10:
                        cur.execute(db.query("create_replication_slot"), {"slot": slot})
                    else:
                        cmd += ["--create-slot"]

        ctx.run(cmd, env=env, check=True)
        for name in keep:
            shutil.copyfile(tmpdir / name, instance.datadir / name)
        # When primary is also managed by pglift, pg_basebackup will also copy
        # conf.pglift.d. So we must drop it to not interfer with site/user
        # config generated by instance_configure
        shutil.rmtree(instance.datadir / "conf.pglift.d", ignore_errors=True)


@task("initialize PostgreSQL instance")
def init(ctx: BaseContext, manifest: interface.Instance) -> None:
    """Initialize a PostgreSQL instance."""
    settings = ctx.settings.postgresql
    initdb_settings = settings.initdb
    surole = settings.surole

    if exists(ctx, manifest.name, manifest.version):
        return None

    instance = BaseInstance.get(manifest.name, manifest.version, ctx)

    # Would raise SystemError if requested postgresql binaries are not
    # available or if versions mismatch.
    pg_ctl = ctx.pg_ctl(manifest.version)

    pgroot = settings.root
    pgroot.parent.mkdir(parents=True, exist_ok=True)
    pgroot.mkdir(mode=0o750, exist_ok=True)

    settings.socket_directory.mkdir(parents=True, exist_ok=True)

    opts: Dict[str, Union[str, Literal[True]]] = {
        "waldir": str(instance.waldir),
        "username": surole.name,
        "encoding": "UTF8",
        # Set temporary auth methods, until the complete pg_hba.conf gets
        # deployed.
        "auth_local": "trust",
        "auth_host": "reject",
    }
    if initdb_settings.locale:
        opts["locale"] = initdb_settings.locale
    if initdb_settings.data_checksums:
        opts["data_checksums"] = True

    pg_ctl.init(instance.datadir, **opts)
    if manifest.standby:
        init_replication(ctx, instance, manifest.standby.for_, manifest.standby.slot)

    if ctx.settings.service_manager == "systemd":
        systemd.enable(ctx, systemd_unit(instance))

    return None


@init.revert("delete PostgreSQL instance")
def revert_init(ctx: BaseContext, manifest: interface.Instance) -> None:
    """Un-initialize a PostgreSQL instance."""
    instance = BaseInstance.get(manifest.name, manifest.version, ctx)
    if ctx.settings.service_manager == "systemd":
        systemd.disable(ctx, systemd_unit(instance), now=True)

    settings = ctx.settings.postgresql
    if instance.path.exists():
        shutil.rmtree(instance.path)
    pgroot = settings.root
    if pgroot.exists():
        try:
            next(pgroot.iterdir())
        except StopIteration:
            # directory is empty
            pgroot.rmdir()


def configure(
    ctx: BaseContext,
    manifest: interface.Instance,
    *,
    ssl: Union[bool, Tuple[Path, Path]] = False,
    **confitems: Optional[pgconf.Value],
) -> ConfigChanges:
    """Write instance's configuration and include it in its postgresql.conf.

    `ssl` parameter controls SSL configuration. If False, SSL is not enabled.
    If True, a self-signed certificate is generated. A tuple of two
    `~pathlib.Path` corresponding to the location of SSL cert file and key
    file to use may also be passed.

    'shared_buffers' and 'effective_cache_size' setting, if defined and set to
    a percent-value, will be converted to proper memory value relative to the
    total memory available on the system.
    """
    instance = PostgreSQLInstance.system_lookup(ctx, (manifest.name, manifest.version))
    configdir = instance.datadir
    postgresql_conf = configdir / "postgresql.conf"
    confd, include = conf.info(configdir)
    if not confd.exists():
        confd.mkdir()
    site_conffile = confd / "site.conf"
    user_conffile = confd / "user.conf"
    pgconfig = pgconf.parse(str(postgresql_conf))
    if ssl:
        confitems["ssl"] = True
    if not pgconfig.get("ssl", False) and ssl is True:
        util.generate_certificate(configdir, run_command=ctx.run)
    elif isinstance(ssl, tuple):
        try:
            certfile, keyfile = ssl
        except ValueError:
            raise ValueError("expecting a 2-tuple for 'ssl' parameter") from None
        confitems["ssl_cert_file"] = str(certfile)
        confitems["ssl_key_file"] = str(keyfile)
    original_content = postgresql_conf.read_text()
    if not any(line.startswith(include) for line in original_content.splitlines()):
        with postgresql_conf.open("w") as f:
            f.write(f"{include}\n\n")
            f.write(original_content)

    site_confitems: Dict[str, Optional[pgconf.Value]] = {"cluster_name": instance.name}
    site_config_template = datapath / "postgresql" / "site.conf"
    if site_config_template.exists():
        site_confitems.update(pgconf.parse(site_config_template).as_dict())

    def format_values(
        confitems: Dict[str, Any], memtotal: float = util.total_memory()
    ) -> None:
        for k in ("shared_buffers", "effective_cache_size"):
            try:
                v = confitems[k]
            except KeyError:
                continue
            if v is None:
                continue
            try:
                confitems[k] = util.percent_memory(v, memtotal)
            except ValueError:
                pass
        for k, v in confitems.items():
            if isinstance(v, str):
                confitems[k] = v.format(settings=ctx.settings.postgresql)

    format_values(confitems)
    format_values(site_confitems)

    def make_config(
        fpath: Path, items: Dict[str, Optional[pgconf.Value]]
    ) -> ConfigChanges:
        config = conf.make(instance.name, **items)

        config_before = {}
        if fpath.exists():
            config_before = {e.name: e.value for e in pgconf.parse(fpath)}
        config_after = {e.name: e.value for e in config}
        changes: ConfigChanges = {}
        for k in set(config_before) | set(config_after):
            pv = config_before.get(k)
            nv = config_after.get(k)
            if nv != pv:
                changes[k] = (pv, nv)

        if changes:
            with fpath.open("w") as f:
                config.save(f)

        return changes

    make_config(site_conffile, site_confitems)
    changes = make_config(user_conffile, confitems)

    i = PostgreSQLInstance.system_lookup(ctx, instance)
    i_config = i.config()
    ctx.pm.hook.instance_configure(
        ctx=ctx, manifest=manifest, config=i_config, changes=changes
    )

    if "log_directory" in i_config:
        logdir = Path(i_config.log_directory)  # type: ignore[arg-type]
        conf.create_log_directory(instance, logdir)

    return changes


@contextlib.contextmanager
def running(
    ctx: BaseContext,
    instance: Union[PostgreSQLInstance, Instance],
    *,
    timeout: int = 10,
    run_hooks: bool = False,
) -> Iterator[None]:
    """Context manager to temporarily start an instance.

    :param timeout: delay to wait for instance startup.
    :param run_hooks: whether or not to run hooks during instance start/stop.

    :raises ~exceptions.InstanceStateError: when the instance did not
        get through running state after specified `timeout` (in seconds).
    """
    if status(ctx, instance) == Status.running:
        yield
        return

    if run_hooks and not isinstance(instance, Instance):
        raise TypeError("expecting a full instance")

    start(ctx, instance, run_hooks=run_hooks)
    for __ in range(timeout):
        time.sleep(1)
        if status(ctx, instance) == Status.running:
            break
    else:
        raise exceptions.InstanceStateError(f"{instance} not started after {timeout}s")
    try:
        yield
    finally:
        stop(ctx, instance, run_hooks=run_hooks)


@contextlib.contextmanager
def stopped(
    ctx: BaseContext,
    instance: Union[PostgreSQLInstance, Instance],
    *,
    timeout: int = 10,
    run_hooks: bool = False,
) -> Iterator[None]:
    """Context manager to temporarily stop an instance.

    :param timeout: delay to wait for instance stop.
    :param run_hooks: whether or not to run hooks during instance start/stop.

    :raises ~exceptions.InstanceStateError: when the instance did stop after
        specified `timeout` (in seconds).
    """
    if status(ctx, instance) == Status.not_running:
        yield
        return

    if run_hooks and not isinstance(instance, Instance):
        raise TypeError("expecting a full instance")

    stop(ctx, instance, run_hooks=run_hooks)
    for __ in range(timeout):
        time.sleep(1)
        if status(ctx, instance) == Status.not_running:
            break
    else:
        raise exceptions.InstanceStateError(f"{instance} not stopped after {timeout}s")
    try:
        yield
    finally:
        start(ctx, instance, run_hooks=run_hooks)


@hookimpl  # type: ignore[misc]
def instance_configure(
    ctx: BaseContext, manifest: interface.Instance, **kwargs: Any
) -> None:
    """Configure authentication for the PostgreSQL instance by setting
    super-user role's password, if any, and installing templated pg_hba.conf
    and pg_ident.conf.

    This is a no-op if if pg_hba.conf's content matches the initial
    configuration.
    """
    surole = interface.instance_surole(ctx.settings, manifest)
    auth_settings = ctx.settings.postgresql.auth
    instance = Instance.system_lookup(ctx, (manifest.name, manifest.version))
    hba_path = instance.datadir / "pg_hba.conf"
    hba = template("postgresql", "pg_hba.conf").format(
        surole=surole.name,
        auth_local=auth_settings.local,
        auth_host=auth_settings.host,
    )
    if hba_path.read_text() == hba:
        return

    if not instance.standby and surole.password:
        # standby instances are read-only
        with running(ctx, instance):
            roles.set_password_for(ctx, instance, surole)

    hba_path.write_text(hba)

    ident_path = instance.datadir / "pg_ident.conf"
    ident = template("postgresql", "pg_ident.conf")
    ident_path.write_text(ident)


def start(
    ctx: BaseContext,
    instance: Union[PostgreSQLInstance, Instance],
    *,
    wait: bool = True,
    logfile: Optional[Path] = None,
    run_hooks: bool = True,
    foreground: bool = False,
) -> None:
    """Start an instance.

    :param run_hooks: controls whether start-up hook will be triggered or not.
    :param foreground: start postgres in the foreground, replacing the current
        process.

    .. note:: When starting in "foreground", hooks will not be triggered and
        `wait` and `logfile` parameters have no effect.
    """
    if run_hooks and not isinstance(instance, Instance):
        raise TypeError("expecting a full instance")

    ctx.info("starting instance %s", instance)
    if foreground and run_hooks:
        ctx.debug("not running hooks for a foreground start")
        run_hooks = False

    start_postgresql(ctx, instance, wait=wait, logfile=logfile, foreground=foreground)

    if run_hooks and wait:
        ctx.pm.hook.instance_start(ctx=ctx, instance=instance)


@task("start PostgreSQL instance")
def start_postgresql(
    ctx: BaseContext,
    instance: Union[PostgreSQLInstance, Instance],
    *,
    wait: bool = True,
    logfile: Optional[Path] = None,
    foreground: bool = False,
) -> None:
    if ctx.settings.service_manager is None:
        if foreground:
            postgres = ctx.pg_ctl(instance.version).bindir / "postgres"
            cmd.execute_program(
                [str(postgres), "-D", str(instance.datadir)], logger=ctx
            )
        else:
            ctx.pg_ctl(instance.version).start(
                instance.datadir, wait=wait, logfile=logfile
            )
    elif ctx.settings.service_manager == "systemd":
        systemd.start(ctx, systemd_unit(instance))


@task("get PostgreSQL instance status")
def status(ctx: BaseContext, instance: BaseInstance) -> Status:
    """Return the status of an instance."""
    return ctx.pg_ctl(instance.version).status(instance.datadir)


def check_status(ctx: BaseContext, instance: BaseInstance, expected: Status) -> None:
    """Check actual instance status with respected to `expected` one.

    :raises ~exceptions.InstanceStateError: in case the actual status is not expected.
    """
    st = status(ctx, instance)
    if st != expected:
        raise exceptions.InstanceStateError(f"instance is {st.name}")


def stop(
    ctx: BaseContext,
    instance: Union[PostgreSQLInstance, Instance],
    *,
    mode: str = "fast",
    wait: bool = True,
    run_hooks: bool = True,
) -> None:
    """Stop an instance.

    :param run_hooks: controls whether stop hook will be triggered or not.
    """
    if run_hooks and not isinstance(instance, Instance):
        raise TypeError("expecting a full instance")

    if status(ctx, instance) == Status.not_running:
        ctx.warning("instance %s is already stopped", instance)
    else:
        stop_postgresql(ctx, instance, mode=mode, wait=wait)
    if run_hooks and wait:
        ctx.pm.hook.instance_stop(ctx=ctx, instance=instance)


@task("stop PostgreSQL instance")
def stop_postgresql(
    ctx: BaseContext,
    instance: Union[PostgreSQLInstance, Instance],
    mode: str = "fast",
    wait: bool = True,
) -> None:
    ctx.info("stopping instance %s", instance)
    if ctx.settings.service_manager is None:
        ctx.pg_ctl(instance.version).stop(instance.datadir, mode=mode, wait=wait)
    elif ctx.settings.service_manager == "systemd":
        systemd.stop(ctx, systemd_unit(instance))


@task("restart PostgreSQL instance")
def restart(
    ctx: BaseContext,
    instance: Instance,
    *,
    mode: str = "fast",
    wait: bool = True,
) -> None:
    """Restart an instance."""
    ctx.info("restarting instance %s", instance)
    if ctx.settings.service_manager is None:
        ctx.pg_ctl(instance.version).restart(instance.datadir, mode=mode, wait=wait)
    elif ctx.settings.service_manager == "systemd":
        systemd.restart(ctx, systemd_unit(instance))


@task("reload PostgreSQL instance")
def reload(
    ctx: BaseContext,
    instance: Instance,
) -> None:
    """Reload an instance."""
    ctx.info("reloading instance %s", instance)
    if ctx.settings.service_manager is None:
        ctx.pg_ctl(instance.version).reload(instance.datadir)
    elif ctx.settings.service_manager == "systemd":
        systemd.reload(ctx, systemd_unit(instance))


def promote(ctx: BaseContext, instance: Instance) -> None:
    """Promote a standby instance"""
    pg_ctl = ctx.pg_ctl(instance.version)
    pg_ctl.run_command(
        [str(pg_ctl.pg_ctl), "promote", "-D", str(instance.datadir)],
        check=True,
    )


def upgrade(
    ctx: BaseContext,
    instance: Instance,
    *,
    version: Optional[str] = None,
    name: Optional[str] = None,
    port: Optional[int] = None,
    jobs: Optional[int] = None,
) -> Instance:
    """Upgrade an instance using pg_upgrade"""
    if version is None:
        version = default_postgresql_version(ctx)
    surole = ctx.settings.postgresql.surole
    surole_password = ctx.libpq_environ().get("PGPASSWORD")
    if not surole_password and ctx.settings.postgresql.auth.passfile:
        with pgpass.edit(ctx.settings.postgresql.auth.passfile) as passfile:
            for entry in passfile:
                if entry.matches(port=instance.port, username=surole.name):
                    surole_password = entry.password
    new_manifest = interface.Instance(
        name=name or instance.name,
        version=version,
        port=port or instance.port,
        state=interface.InstanceState.stopped,
        prometheus={"port": instance.prometheus.port},
        surole_password=SecretStr(surole_password) if surole_password else None,
    )
    result = apply(ctx, new_manifest)
    assert result is not None, new_manifest
    (newinstance, _) = result
    bindir = ctx.pg_ctl(version).bindir
    pg_upgrade = str(bindir / "pg_upgrade")
    cmd = [
        pg_upgrade,
        f"--old-bindir={ctx.pg_ctl(instance.version).bindir}",
        f"--new-bindir={bindir}",
        f"--old-datadir={instance.datadir}",
        f"--new-datadir={newinstance.datadir}",
        f"--username={ctx.settings.postgresql.surole.name}",
    ]
    if jobs is not None:
        cmd.extend(["--jobs", str(jobs)])
    env = ctx.libpq_environ()
    if surole_password:
        env.setdefault("PGPASSWORD", surole_password)
    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            ctx.run(cmd, check=True, cwd=tmpdir, env=env)
    except exceptions.CommandError:
        drop(ctx, newinstance)
        raise
    return newinstance


def apply(
    ctx: BaseContext, manifest: interface.Instance
) -> Optional[Tuple[Instance, ConfigChanges]]:
    """Apply state described by specified manifest as a PostgreSQL instance.

    Depending on the previous state and existence of the target instance, the
    instance may be created or updated or dropped.

    Unless the target state is 'absent', return an
    :class:`~pglift.model.Instance` object along with configuration changes.

    If configuration changes are detected and the instance was previously
    running, it will be reloaded. Note that some changes require a full
    restart, this needs to be handled manually.
    """
    States = interface.InstanceState
    state = manifest.state

    if state == States.absent:
        if exists(ctx, manifest.name, manifest.version):
            drop(ctx, Instance.system_lookup(ctx, (manifest.name, manifest.version)))
        return None

    if not exists(ctx, manifest.name, manifest.version):
        init(ctx, manifest)

    configure_options = manifest.configuration or {}
    changes = configure(
        ctx,
        manifest,
        ssl=manifest.ssl,
        port=manifest.port,
        **configure_options,
    )

    instance = Instance.system_lookup(ctx, (manifest.name, manifest.version))
    is_running = status(ctx, instance) == Status.running
    if state == States.stopped:
        if is_running:
            stop(ctx, instance)
    elif state == States.started:
        if is_running:
            if changes:
                # This might fail because some changes require a restart but
                # 'pg_ctl reload' does not inform us about that.
                reload(ctx, instance)
        else:
            start(ctx, instance)
    else:
        assert False, f"unexpected state: {state}"  # pragma: nocover

    StandbyState = manifest.Standby.State

    if (
        manifest.standby
        and manifest.standby.status == StandbyState.promoted
        and instance.standby is not None
    ):
        promote(ctx, instance)

    return instance, changes


def describe(ctx: BaseContext, name: str, version: Optional[str]) -> interface.Instance:
    """Return an instance described as a manifest."""
    instance = Instance.system_lookup(ctx, (name, version))
    config = instance.config()
    managed_config = instance.config(managed_only=True).as_dict()
    managed_config.pop("port", None)
    state = interface.InstanceState.from_pg_status(status(ctx, instance))
    prometheus = interface.Instance.Prometheus(port=instance.prometheus.port)
    return interface.Instance(
        name=instance.name,
        version=instance.version,
        port=instance.port,
        state=state,
        ssl=config.ssl,
        configuration=managed_config,
        prometheus=prometheus,
    )


@task("drop PostgreSQL instance")
def drop(ctx: BaseContext, instance: Instance) -> None:
    """Drop an instance."""
    stop(ctx, instance, run_hooks=True)

    ctx.pm.hook.instance_drop(ctx=ctx, instance=instance)
    manifest = interface.Instance(name=instance.name, version=instance.version)
    revert_init(ctx, manifest)


def list(
    ctx: BaseContext, *, version: Optional[str] = None
) -> Iterator[interface.InstanceListItem]:
    """Yield instances found by system lookup.

    :param version: filter instances matching a given version.

    :raises ~exceptions.InvalidVersion: if specified version is unknown.
    """
    versions = builtins.list(ctx.settings.postgresql.versions)
    if version:
        if version not in versions:
            raise exceptions.InvalidVersion(f"unknown version '{version}'")
        versions = [version]

    pgroot = ctx.settings.postgresql.root

    # Search for directories looking like <version>/<name> in pgroot
    for ver in versions:
        version_path = pgroot / ver
        if not version_path.is_dir():
            continue
        for d in version_path.iterdir():
            if not d.is_dir():
                continue
            try:
                instance = Instance.system_lookup(ctx, (d.name, ver))
            except exceptions.InstanceNotFound:
                continue

            yield interface.InstanceListItem(
                name=instance.name,
                path=str(instance.path),
                port=instance.port,
                status=status(ctx, instance).name,
                version=instance.version,
            )


def shell(
    ctx: BaseContext, instance: Instance, *, user: Optional[str], dbname: Optional[str]
) -> None:
    """Start a PostgreSQL shell (psql) and replace the current process.

    The instance should be running.
    """
    config = instance.config()
    try:
        host = config.unix_socket_directories.split(",")[0]  # type: ignore[union-attr]
    except (AttributeError, IndexError):
        host = "localhost"
    if user is None:
        user = ctx.settings.postgresql.surole.name
    args = [
        str(ctx.pg_ctl(instance.version).bindir / "psql"),
        "--port",
        str(instance.port),
        "--host",
        host,
        "--user",
        user,
    ]
    if dbname is not None:
        args.extend(["--dbname", dbname])
    env = ctx.libpq_environ()
    cmd.execute_program(args, logger=ctx, env=env)


def exists(ctx: BaseContext, name: str, version: Optional[str]) -> bool:
    """Return true when instance exists"""
    try:
        PostgreSQLInstance.system_lookup(ctx, (name, version))
    except exceptions.InstanceNotFound:
        return False
    return True



import sys
from typing import Optional

from . import systemd
from .ctx import BaseContext
from .settings import Settings
from .task import task

POSTGRESQL_SERVICE_NAME = "pglift-postgresql@.service"
POSTGRES_EXPORTER_SERVICE_NAME = "pglift-postgres_exporter@.service"
BACKUP_SERVICE_NAME = "pglift-backup@.service"
BACKUP_TIMER_NAME = "pglift-backup@.timer"


def with_header(content: str, header: str) -> str:
    """Possibly insert `header` on top of `content`.

    >>> print(with_header("blah", "% head"))
    % head
    blah
    >>> with_header("content", "")
    'content'
    """
    if header:
        content = "\n".join([header, content])
    return content


def executeas(settings: Settings) -> str:
    """Return User/Group options for systemd unit depending on settings.

    When systemd.user is set, return an empty string:
    >>> s = Settings.parse_obj({"systemd": {"user": True}})
    >>> executeas(s)
    ''

    Otherwise, user sysuser setting:
    >>> s = Settings.parse_obj({"systemd": {"user": False}, "sysuser": ["postgres", "pgsql"]})
    >>> print(executeas(s))
    User=postgres
    Group=pgsql
    """
    if settings.systemd.user:
        return ""
    user, group = settings.sysuser
    return "\n".join([f"User={user}", f"Group={group}"])


@task("install systemd template unit for PostgreSQL")
def postgresql_systemd_unit_template(
    ctx: BaseContext, *, env: Optional[str] = None, header: str = ""
) -> None:
    settings = ctx.settings.postgresql
    environment = ""
    if env:
        environment = f"\nEnvironment={env}\n"
    content = systemd.template(POSTGRESQL_SERVICE_NAME).format(
        executeas=executeas(ctx.settings),
        python=sys.executable,
        environment=environment,
        pid_directory=settings.pid_directory,
    )
    systemd.install(
        POSTGRESQL_SERVICE_NAME,
        with_header(content, header),
        ctx.settings.systemd.unit_path,
        logger=ctx,
    )


@postgresql_systemd_unit_template.revert(
    "uninstall systemd template unit for PostgreSQL"
)
def revert_postgresql_systemd_unit_template(
    ctx: BaseContext, *, env: Optional[str] = None, header: str = ""
) -> None:
    systemd.uninstall(
        POSTGRESQL_SERVICE_NAME, ctx.settings.systemd.unit_path, logger=ctx
    )


@task("install systemd template unit for Prometheus postgres_exporter")
def postgres_exporter_systemd_unit_template(
    ctx: BaseContext, *, header: str = ""
) -> None:
    settings = ctx.settings.prometheus
    configpath = str(settings.configpath).replace("{name}", "%i")
    content = systemd.template(POSTGRES_EXPORTER_SERVICE_NAME).format(
        executeas=executeas(ctx.settings),
        configpath=configpath,
        execpath=settings.execpath,
    )
    systemd.install(
        POSTGRES_EXPORTER_SERVICE_NAME,
        with_header(content, header),
        ctx.settings.systemd.unit_path,
        logger=ctx,
    )


@postgres_exporter_systemd_unit_template.revert(
    "uninstall systemd template unit for Prometheus postgres_exporter"
)
def revert_postgres_exporter_systemd_unit_template(
    ctx: BaseContext, *, header: str = ""
) -> None:
    systemd.uninstall(
        POSTGRES_EXPORTER_SERVICE_NAME, ctx.settings.systemd.unit_path, logger=ctx
    )


@task("install systemd template unit and timer for PostgreSQL backups")
def postgresql_backup_systemd_templates(
    ctx: BaseContext, *, env: Optional[str] = None, header: str = ""
) -> None:
    environment = ""
    if env:
        environment = f"\nEnvironment={env}\n"
    service_content = systemd.template(BACKUP_SERVICE_NAME).format(
        executeas=executeas(ctx.settings),
        environment=environment,
        python=sys.executable,
    )
    systemd.install(
        BACKUP_SERVICE_NAME,
        with_header(service_content, header),
        ctx.settings.systemd.unit_path,
        logger=ctx,
    )
    timer_content = systemd.template(BACKUP_TIMER_NAME).format(
        # TODO: use a setting for that value
        calendar="daily",
    )
    systemd.install(
        BACKUP_TIMER_NAME,
        with_header(timer_content, header),
        ctx.settings.systemd.unit_path,
        logger=ctx,
    )


@postgresql_backup_systemd_templates.revert(
    "uninstall systemd template unit and timer for PostgreSQL backups"
)
def revert_postgresql_backup_systemd_templates(
    ctx: BaseContext, *, env: Optional[str] = None, header: str = ""
) -> None:
    systemd.uninstall(BACKUP_SERVICE_NAME, ctx.settings.systemd.unit_path, logger=ctx)
    systemd.uninstall(BACKUP_TIMER_NAME, ctx.settings.systemd.unit_path, logger=ctx)


def do(ctx: BaseContext, env: Optional[str] = None, header: str = "") -> None:
    if ctx.settings.service_manager != "systemd":
        ctx.warning("not using systemd as 'service_manager', skipping installation")
        return
    postgresql_systemd_unit_template(ctx, env=env, header=header)
    postgres_exporter_systemd_unit_template(ctx, header=header)
    postgresql_backup_systemd_templates(ctx, env=env, header=header)
    systemd.daemon_reload(ctx)


def undo(ctx: BaseContext) -> None:
    if ctx.settings.service_manager != "systemd":
        ctx.warning("not using systemd as 'service_manager', skipping uninstallation")
        return
    revert_postgresql_backup_systemd_templates(ctx)
    revert_postgres_exporter_systemd_unit_template(ctx)
    revert_postgresql_systemd_unit_template(ctx)
    systemd.daemon_reload(ctx)



from contextlib import contextmanager
from functools import partial
from typing import Any, Iterator, List, Optional, overload

from typing_extensions import Literal

from pglift import db
from pglift import instance as instance_mod
from pglift.ctx import BaseContext
from pglift.models import interface
from pglift.models.system import Instance
from pglift.types import Role


def configure_instance(
    ctx: BaseContext,
    manifest: interface.Instance,
    *,
    port: Optional[int] = None,
    **confitems: Any,
) -> None:
    if port is None:
        port = manifest.port
    instance_mod.configure(ctx, manifest, port=port, **confitems)


@contextmanager
def reconfigure_instance(
    ctx: BaseContext, i: Instance, manifest: interface.Instance, *, port: int
) -> Iterator[None]:
    config = i.config()
    assert config is not None
    initial_port = config.port
    assert initial_port
    configure_instance(ctx, manifest, port=port)
    try:
        yield
    finally:
        configure_instance(ctx, manifest, port=initial_port)  # type: ignore[arg-type]


@overload
def execute(
    ctx: BaseContext,
    instance: Instance,
    query: str,
    fetch: Literal[True],
    autocommit: bool = False,
    role: Optional[Role] = None,
    **kwargs: Any,
) -> List[Any]:
    ...


@overload
def execute(
    ctx: BaseContext,
    instance: Instance,
    query: str,
    fetch: bool = False,
    autocommit: bool = False,
    role: Optional[Role] = None,
    **kwargs: Any,
) -> List[Any]:
    ...


def execute(
    ctx: BaseContext,
    instance: Instance,
    query: str,
    fetch: bool = True,
    autocommit: bool = False,
    role: Optional[Role] = None,
    **kwargs: Any,
) -> Optional[List[Any]]:
    if role is None:
        connect = partial(db.superuser_connect, ctx)
    elif role.password:
        connect = partial(
            db.connect, user=role.name, password=role.password.get_secret_value()
        )
    else:
        connect = partial(db.connect, user=role.name)
    with instance_mod.running(ctx, instance):
        with connect(instance, autocommit=autocommit, **kwargs) as conn:
            with conn.cursor() as cur:
                cur.execute(query)
                conn.commit()
                if fetch:
                    return cur.fetchall()  # type: ignore[no-any-return]
        return None



import subprocess
from typing import Any, List, NoReturn

import pytest

from pglift import postgres
from pglift.ctx import Context
from pglift.models.system import Instance


def test_main_errors() -> None:
    with pytest.raises(SystemExit, match="2"):
        postgres.main(["aa"])
    with pytest.raises(SystemExit, match="2"):
        postgres.main(["12-"])
    with pytest.raises(SystemExit, match="2"):
        postgres.main(["12-test"])


def test_main(
    monkeypatch: pytest.MonkeyPatch, ctx: Context, instance: Instance
) -> None:
    calls = []

    class Popen:
        def __init__(self, cmd: List[str], **kwargs: Any):
            calls.append(cmd)
            self.cmd = cmd
            self.pid = 123
            self.returncode = 0

        def communicate(self, **kwargs: Any) -> NoReturn:
            raise subprocess.TimeoutExpired(self.cmd, 12)

    with monkeypatch.context() as m:
        m.setattr("subprocess.Popen", Popen)
        postgres.main([f"{instance.version}-{instance.name}"], ctx=ctx)
    bindir = ctx.settings.postgresql.versions[instance.version].bindir
    assert calls == [[str(bindir / "postgres"), "-D", str(instance.datadir)]]
    assert (
        ctx.settings.postgresql.pid_directory
        / f"postgresql-{instance.version}-{instance.name}.pid"
    ).read_text() == "123"
